@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2009-2011, 2013, 2017, 2019 Sergey Poznyakoff
+   Copyright (C) 2009-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,9 +14,6 @@
    You should have received a copy of the GNU General Public License
    along with GNU Pies.  If not, see <http://www.gnu.org/licenses/>. */
 
-int assert_grecs_value_type (grecs_locus_t *locus,
-			     const grecs_value_t *value, int type);
-int assert_scalar_stmt (grecs_locus_t *locus, enum grecs_callback_command cmd);
 int conf_callback_url (enum grecs_callback_command cmd,
-		       grecs_locus_t *locus,
-		       void *varptr, grecs_value_t *value, void *cb_data);
+		       grecs_node_t *node,
+		       void *varptr, void *cb_data);
