@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2016-2019 Sergey Poznyakoff
+   Copyright (C) 2016-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -154,7 +154,6 @@ component_create (const char *name)
     {
       comp = grecs_zalloc (sizeof (*comp));
       comp->listidx = cur;
-      comp->facility = log_facility;
       comp->redir[RETR_OUT].type = comp->redir[RETR_ERR].type = redir_null;
       comp->tag = grecs_strdup (name);
       comp->socket_type = SOCK_STREAM;
@@ -293,7 +292,6 @@ component_match (struct component *comp, struct component *ref)
   EQ (pass_fd_timeout);
   FN (acl, pies_acl_cmp);
   FN (tcpmux, safe_strcmp);
-  EQ (facility);
   FNP (redir[0], redirector_cmp);
   FNP (redir[1], redirector_cmp);
 #undef MATCH
@@ -403,44 +401,24 @@ list_str_cmp (const void *a, const void *b)
    DP is a transitive closure of depmap.
 */
 static void
-report_cyclic_dependency (pies_depmap_t dp, size_t idx)
+report_cyclic_dependency (struct depmap_path *path)
 {
-  size_t i;
-
-  i = idx;
-  do
+  struct depmap_path_elem *p;
+  for (p = path->head; p; p = p->next)
     {
-      size_t n;
-      pies_depmap_pos_t pos;
-
-      logmsg_printf (LOG_NOTICE, "%s -> ", comp_array[i]->tag);
-      comp_array[i]->flags |= CF_REMOVE;
-      for (n = depmap_first (depmap, depmap_col, i, &pos);
-	   n != (size_t)-1;
-	   n = depmap_next (depmap, pos))
-	{
-	  if (n == i)
-	    continue;
-	  if (depmap_isset (dp, n, i) && depmap_isset (dp, n, n))
-	    break;
-	}
-      depmap_end (pos);
-
-      if (n == (size_t)-1)
-	break;
-      i = n;
+      logmsg_printf (LOG_NOTICE, "%s -> ", comp_array[p->idx]->tag);
+      comp_array[p->idx]->flags |= CF_REMOVE;
     }
-  while (i != idx);
-  logmsg_printf (LOG_NOTICE, "%s\n", comp_array[idx]->tag);
+  logmsg_printf (LOG_NOTICE, "%s\n", comp_array[path->head->idx]->tag);
 }
 
 void
 component_build_depmap (void)
 {
-  size_t i;
-  pies_depmap_t dp;
-
-  free (depmap);
+  struct depmap_path *path;
+  int i;
+  
+  depmap_free (depmap);
   depmap = depmap_alloc (comp_count);
   for (i = 0; i < comp_count; )
     {
@@ -482,24 +460,22 @@ component_build_depmap (void)
       i++;
     }
 
-  dp = depmap_copy (depmap);
-  depmap_tc (dp);
-  for (i = 0; i < comp_count; i++)
-    if (!(comp_array[i]->flags & CF_REMOVE) && depmap_isset (dp, i, i))
-      {
-	logmsg (LOG_ERR, _("component %s depends on itself"),
-		comp_array[i]->tag);
-	report_cyclic_dependency (dp, i);
-      }
-
-
-  for (i = 0; i < comp_count;)
-    if (comp_array[i]->flags & CF_REMOVE)
-      comp_array_remove (i);
-    else
-      i++;
-
-  free (dp);
+  path = depmap_cycle_detect (depmap);
+  if (path)
+    {
+      struct depmap_path *p;
+      
+      logmsg (LOG_ERR, "%s", _("cyclic dependencies detected:"));
+      for (p = path; p; p = p->next)
+	report_cyclic_dependency (p);
+      depmap_path_free (path);
+      
+      for (i = 0; i < comp_count;)
+	if (comp_array[i]->flags & CF_REMOVE)
+	  comp_array_remove (i);
+	else
+	  i++;
+    }
 }
 
 void
@@ -746,6 +722,14 @@ component_finish (struct component *comp, grecs_locus_t *locus)
 
   if (comp->flags & CF_SHELL)
     {
+      if (comp->flags & CF_EXPANDENV)
+	{
+	  grecs_warning (locus, 0,
+			 "%s",
+			 _("flags \"shell\" and \"expandenv\" used together: "
+			   "ignoring \"expandenv\""));
+	  comp->flags &= ~CF_EXPANDENV;
+	}
       comp->argc = 3;
       comp->argv = grecs_calloc (comp->argc + 1, sizeof (comp->argv[0]));
       comp->argv[0] = grecs_strdup (comp->program ? comp->program : "/bin/sh");
@@ -753,7 +737,7 @@ component_finish (struct component *comp, grecs_locus_t *locus)
       comp->argv[2] = grecs_strdup (comp->command);
       comp->argv[3] = NULL;
     }
-  else if (comp->command)
+  else if (comp->command && !(comp->flags & CF_EXPANDENV))
     {
       struct wordsplit ws;
       if (wordsplit (comp->command, &ws, WRDSF_DEFFLAGS))
@@ -806,6 +790,12 @@ depmap_dump (pies_depmap_t dpm)
 {
   size_t i, j;
 
+  if (depmap_dim (dpm) == 0)
+    {
+      printf ("%s\n", _("No components defined"));
+      return;
+    }
+  
   printf ("%s:\n", _("Dependency map"));
   printf ("  ");
   for (i = 0; i < comp_count; i++)
