@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2008-2019 Sergey Poznyakoff
+   Copyright (C) 2008-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,8 +23,6 @@
 int preprocess_only;		/* Preprocess config, do nothing more */
 int lint_mode;			/* Test configuration syntax and exit */
 int log_to_stderr_only;                /* Use only stderr for logging */
-int log_facility = LOG_USER;
-char *log_tag;
 struct pies_privs pies_privs;
 int foreground;
 int init_process;
@@ -62,6 +60,8 @@ char *default_control_url[2] = {
   DEFAULT_PIES_CONTROL_URL,
   DEFAULT_INIT_CONTROL_URL
 };
+
+static envop_t *pies_envop;
 
 struct config_syntax
 {
@@ -707,19 +707,19 @@ cb_env_section_parser (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
 
   switch (cmd)
     {
     case grecs_callback_section_begin:
-      *(struct component **) cb_data = comp;
+      *(envop_t ***) cb_data = envop_ptr;
       break;
 
     case grecs_callback_section_end:
       break;
 
     case grecs_callback_set_value:
-      return _cb_env (&comp->envop, value, locus);
+      return _cb_env (envop_ptr, value, locus);
     }
   return 0;
 }
@@ -731,7 +731,7 @@ _cb_env_clear (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
 
   if (!GRECS_VALUE_EMPTY_P (value))
     {
@@ -739,7 +739,7 @@ _cb_env_clear (enum grecs_callback_command cmd,
       return 1;
     }
   
-  if (envop_entry_add (&comp->envop, envop_clear, NULL, NULL))
+  if (envop_entry_add (envop_ptr, envop_clear, NULL, NULL))
     grecs_error (locus, errno, "envop_entry_add");
 
   return 0;
@@ -752,7 +752,7 @@ _cb_env_keep (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
   char *p;
 
   if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
@@ -760,9 +760,9 @@ _cb_env_keep (enum grecs_callback_command cmd,
   p = strchr (value->v.string, '=');
   if (p)
     *p++ = 0;
-  if (envop_entry_add (&comp->envop, envop_clear, NULL, NULL))
+  if (envop_entry_add (envop_ptr, envop_clear, NULL, NULL))
     grecs_error (locus, errno, "envop_entry_add");
-  if (envop_entry_add (&comp->envop, envop_keep, value->v.string, p))
+  if (envop_entry_add (envop_ptr, envop_keep, value->v.string, p))
     grecs_error (locus, errno, "envop_entry_add");
   return 0;
 }
@@ -774,7 +774,7 @@ _cb_env_set (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
   char *p;
 
   if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
@@ -782,7 +782,7 @@ _cb_env_set (enum grecs_callback_command cmd,
   p = strchr (value->v.string, '=');
   if (p)
     *p++ = 0;
-  if (envop_entry_add (&comp->envop, envop_set, value->v.string, p))
+  if (envop_entry_add (envop_ptr, envop_set, value->v.string, p))
     grecs_error (locus, errno, "envop_entry_add");
   return 0;
 }
@@ -794,11 +794,11 @@ _cb_env_eval (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
 
   if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
     return 1;
-  if (envop_entry_add (&comp->envop, envop_set, NULL, value->v.string))
+  if (envop_entry_add (envop_ptr, envop_set, NULL, value->v.string))
     grecs_error (locus, errno, "envop_entry_add");
   return 0;
 }
@@ -810,7 +810,7 @@ _cb_env_unset (enum grecs_callback_command cmd,
 {
   grecs_locus_t *locus = &node->locus;
   grecs_value_t *value = node->v.value;
-  struct component *comp = varptr;
+  envop_t **envop_ptr = varptr;
   char *p;
 
   if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
@@ -818,7 +818,7 @@ _cb_env_unset (enum grecs_callback_command cmd,
   p = strchr (value->v.string, '=');
   if (p)
     *p++ = 0;
-  if (envop_entry_add (&comp->envop, envop_unset, value->v.string, p))
+  if (envop_entry_add (envop_ptr, envop_unset, value->v.string, p))
     grecs_error (locus, errno, "envop_entry_add");
   return 0;
 }
@@ -878,7 +878,7 @@ string_to_syslog_priority (const char *key, int *pres)
 }
 
 int
-string_to_syslog_facility (const char *key, int *pres)
+string_to_syslog_facility (const char *key, int len, int *pres)
 {
   static struct tokendef tokdef_fac[] = {
     {"auth", LOG_AUTH},
@@ -908,7 +908,52 @@ string_to_syslog_facility (const char *key, int *pres)
     {NULL}
   };
 
-  return strtotok_ci (tokdef_fac, key, pres);
+  return len == 0
+	  ? strtotok_ci (tokdef_fac, key, pres)
+	  : strtotok_len_ci (tokdef_fac, key, len, pres);
+}
+
+static int
+string_to_syslog_fp (const char *str, grecs_locus_t *locus, int *pres)
+{
+  size_t len = strcspn (str, ".");
+  int f = pies_log_facility, p = LOG_ERR;
+
+  if (string_to_syslog_facility (str, len, &f) == 0)
+    {
+      if (str[len])
+	{
+	  if (string_to_syslog_priority (str + len + 1, &p))
+	    {
+	      grecs_error (locus, 0, "%s", _("unknown syslog priority"));
+	      return 1;
+	    }
+	}
+    }
+  else if (str[len])
+    {
+      grecs_error (locus, 0, "%s", _("unknown syslog facility"));
+      return 1;
+    }
+  else if (string_to_syslog_priority (str, &p))
+    {
+      grecs_error (locus, 0, "%s", _("unknown syslog priority"));
+      return 1;
+    }
+  *pres = f | p;
+  return 0;
+}
+
+static int
+cb_syslog_dev (enum grecs_callback_command cmd,
+		    grecs_node_t *node,
+		    void *varptr, void *cb_data)
+{
+  grecs_value_t *value = node->v.value;
+
+  if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
+    return 1;
+  return pies_syslog_set_dev (value->v.string);
 }
 
 static int
@@ -933,7 +978,7 @@ cb_syslog_facility (enum grecs_callback_command cmd,
       else
 	*(int *) varptr = n;
     }
-  else if (string_to_syslog_facility (str, varptr))
+  else if (string_to_syslog_facility (str, 0, varptr))
     grecs_error (locus, 0, _("unknown syslog facility %s"), str);
   return 0;
 }
@@ -963,12 +1008,8 @@ _cb_redir (enum grecs_callback_command cmd,
 	  break;
 	}
       rp->type = redir_syslog;
-      if (string_to_syslog_priority (value->v.string, &rp->v.prio))
-	{
-	  grecs_error (locus, 0, _("unknown syslog priority %s"),
-		       value->v.string);
-	  return 1;
-	}
+      if (string_to_syslog_fp (value->v.string, locus, &rp->v.prio))
+	return 1;
       break;
 
     case GRECS_TYPE_ARRAY:
@@ -997,14 +1038,9 @@ _cb_redir (enum grecs_callback_command cmd,
 		  break;
 
 		case redir_syslog:
-		  if (string_to_syslog_priority (value->v.arg.v[1]->v.string,
-						 &rp->v.prio))
-		    {
-		      grecs_error (locus, 0,
-				   _("unknown syslog priority %s"),
-				   value->v.arg.v[1]->v.string);
-		      return 1;
-		    }
+		  if (string_to_syslog_fp (value->v.arg.v[1]->v.string,
+					   locus, &rp->v.prio))
+		    return 1;
 		  break;
 
 		case redir_file:
@@ -1143,6 +1179,7 @@ str_to_cf (const char *string, int *flags)
     { "siggroup", CF_SIGGROUP },
     { "nullinput", CF_NULLINPUT },
     { "shell", CF_SHELL },
+    { "expandenv", CF_EXPANDENV },
     { NULL }
   };
 
@@ -1344,13 +1381,6 @@ struct grecs_keyword component_keywords[] = {
    NULL, offsetof (struct component, rmfile),
    NULL,
    },
-  {"facility",
-   N_("arg"),
-   N_("Override default syslog facility for this component."),
-   grecs_type_string, GRECS_DFLT,
-   NULL, offsetof (struct component, facility),
-   cb_syslog_facility,
-   },
   {"stdout",
    /* TRANSLATORS: file and syslog are keywords. Do not translate them. */
    N_("type: {file | syslog}> <channel: string"),
@@ -1408,7 +1438,7 @@ struct grecs_keyword component_keywords[] = {
    NULL,
    N_("Modify program environment."),
    grecs_type_section, GRECS_DFLT,
-   NULL, 0,
+   NULL, offsetof (struct component, envop),
    cb_env_section_parser, NULL, cb_env_keywords 
    },
   {"env",
@@ -1416,7 +1446,7 @@ struct grecs_keyword component_keywords[] = {
    N_("Modify program environment (legacy syntax).\n"
       "Argument is a list of quoted assignments separated by white space."),
    grecs_type_string, GRECS_DFLT,
-   NULL, 0,
+   NULL, offsetof (struct component, envop),
    cb_env_section_parser, NULL, NULL
    },
   {"chdir",
@@ -1536,16 +1566,22 @@ struct grecs_keyword control_keywords[] = {
 
 /* syslog */
 static struct grecs_keyword syslog_kw[] = {
+  {"dev",
+   N_("name"),
+   N_("Syslog device: either absolute pathname of the socket file, "
+      "or an IPv4 address and optional port, separated with a colon"),
+   grecs_type_string, GRECS_CONST,
+   NULL, 0, cb_syslog_dev },   
   {"facility",
    N_("name"),
    N_("Set syslog facility. Arg is one of the following: user, daemon, "
       "auth, authpriv, mail, cron, local0 through local7 (case-insensitive), "
       "or a facility number."),
    grecs_type_string, GRECS_DFLT,
-   &log_facility, 0, cb_syslog_facility},
+   &pies_log_facility, 0, cb_syslog_facility},
   {"tag", N_("string"), N_("Tag syslog messages with this string"),
    grecs_type_string, GRECS_DFLT,
-   &log_tag},
+   &pies_log_tag},
 #if 0
   /* This is reserved for future use */
   {
@@ -1589,6 +1625,13 @@ struct grecs_keyword pies_keywords[] = {
    grecs_type_section, GRECS_DFLT,
    NULL, 0,
    component_section_parser, NULL, component_keywords},
+  {"env",
+   NULL,
+   N_("Modify program environment."),
+   grecs_type_section, GRECS_DFLT,
+   &pies_envop, 0,
+   cb_env_section_parser, NULL, cb_env_keywords 
+   },
   {"control",
    NULL,
    N_("Define control socket"),
@@ -1885,6 +1928,13 @@ sig_handler (int sig)
     }
 }
 
+RETSIGTYPE
+sigchld_early (int sig)
+{
+  while (waitpid (-1, NULL, WNOHANG) != -1)
+    ;
+}
+
 void
 setsigvhan (RETSIGTYPE (*handler) (int signo), int *sigv, int sigc)
 {
@@ -2255,7 +2305,66 @@ set_state_file_names (const char *base)
   if (!qotdfile)
     qotdfile = mkfilename (statedir, base, ".qotd");
 }
+
+/* Return 1 if pies is run from docker and 0 otherwise. */
+static int
+is_docker (void)
+{
+  FILE *fp;
+  char *id = NULL;
+  int res;
+  
+  fp = fopen ("/proc/self/cgroup", "r");
+  if (!fp)
+    return 0;
+  res = fscanf (fp, "%*d:%*[^:]:/docker/%ms\n", &id) == 1;
+  fclose (fp);
+  free (id);
+  return res;
+}
 
+/*
+ * Check if `--no-init' option is present in argv/argc.  Return 1 if so,
+ * 0 otherwise.
+ */
+static int
+no_init_option (int argc, char **argv)
+{
+  int i;
+  for (i = 1; i < argc; i++)
+    {
+      if (strcmp (argv[i], "--no-init") == 0)
+	return 1;
+    }
+  return 0;
+}
+
+/*
+ * init_detect - detect if we're run as init process.
+ * 
+ * Pies runs as init process if (1) it has PID 1, (2) is not started as docker
+ * entrypoint, and (3) the `--no-init' command line option is not given.
+ *
+ * The function sets init_process to 1 if pies runs as init process and 0
+ * otherwise.
+ *
+ * No matter the result, if PID is 1 it installs an early SIGCHLD handler,
+ * to ensure that the configuration preprocessor (if such is started) will be
+ * cleaned up properly on exit.
+ */
+static void
+init_detect (int argc, char **argv)
+{
+  init_process = getpid () == 1;
+  if (init_process)
+    {
+      int s[] = { SIGCHLD };
+      setsigvhan (sigchld_early, s, 1);
+      if (no_init_option (argc, argv) || is_docker ())
+	init_process = 0;
+    }
+}
+
 size_t pies_master_argc;
 char **pies_master_argv;
 
@@ -2266,7 +2375,6 @@ main (int argc, char **argv)
   extern char **environ;
   struct grecs_list_entry *ep;
   int diag_flags;
-  int i;
 
   set_program_name (argv[0]);
 #ifdef ENABLE_NLS
@@ -2283,20 +2391,12 @@ main (int argc, char **argv)
 
   set_quoting_style (NULL, shell_quoting_style);
 
-  init_process = getpid () == 1;
-  for (i = 1; i < argc; i++)
-    {
-      if (strcmp (argv[i], "--no-init") == 0)
-	{
-	  init_process = 0;
-	  break;
-	}
-    }
-
+  init_detect (argc, argv);
+  
   /* Set default logging */
   if (SYSVINIT_ACTIVE)
     {
-      log_facility = LOG_DAEMON;
+      pies_log_facility = LOG_DAEMON;
       diag_flags = DIAG_TO_STDERR | DIAG_REOPEN_LOG;
     }
   else
@@ -2325,7 +2425,7 @@ main (int argc, char **argv)
 	instance++;
     }
   setenv ("PIES_INSTANCE", instance, 1);
-  log_tag = grecs_strdup (instance);
+  pies_log_tag = grecs_strdup (instance);
 
   set_conf_file_names (instance);
 
@@ -2354,6 +2454,32 @@ main (int argc, char **argv)
   set_state_file_names (instance);
   set_mailer_argcv ();
 
+  if (pies_envop)
+    {
+      environ_t *env;
+      
+      if ((env = environ_create (environ)) == NULL)
+	{
+	  logmsg (LOG_CRIT, "environ_create: %s", strerror (errno));
+	  exit (EX_OSERR);
+	}
+      if (envop_exec (pies_envop, env))
+	{
+	  logmsg (LOG_CRIT, "environ_exec: %s", strerror (errno));
+	  exit (EX_OSERR);
+	}
+      environ = environ_ptr (env);
+
+      if (debug_level >= 4)
+	{
+	  int i;
+	  logmsg_printf (LOG_DEBUG, "environment: ");
+	  for (i = 0; environ[i]; i++)
+	    logmsg_printf (LOG_DEBUG, "%s ", environ[i]);
+	  logmsg_printf (LOG_DEBUG, "\n");
+	}
+    }
+  
   if (lint_mode)
     exit (0);
 
@@ -2548,14 +2674,8 @@ main (int argc, char **argv)
 
   if (action == ACTION_RESTART)
     {
-      int minfd = DIAG_OUTPUT (DIAG_TO_STDERR) ? 2 : 0;
-      int i;
-
-      for (i = getmaxfd (); i > minfd; i--)
-	close (i);
-
+      pies_close_fds (3);
       signal_setup (SIG_DFL);
-
       execv (pies_master_argv[0], pies_master_argv);
     }
 
