@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2008-2019 Sergey Poznyakoff
+   Copyright (C) 2008-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -58,6 +58,7 @@
 #include "libpies.h"
 #include "envop.h"
 #include "grecs/json.h"
+#include "pies_syslog.h"
 
 #define ARRAY_SIZE(a) (sizeof(a)/sizeof((a)[0]))
 
@@ -209,6 +210,8 @@ enum pies_comp_mode
 #define CF_SIGGROUP    0x100 /* Send signals to the process group */
 #define CF_NULLINPUT   0x200 /* Provide null input stream */
 #define CF_SHELL       0x400 /* Invoke via sh -c */
+#define CF_EXPANDENV   0x800 /* Expand environment variables in the command
+				line */
 
 #define CF_REMOVE      0xf000 /* Marked for removal */
 
@@ -267,7 +270,6 @@ struct component
   char *max_ip_connections_message;
 
   /* Redirectors: */
-  int facility;		 /* Syslog facility. */
   struct redirector redir[2];   /* Repeaters for stdout and stderr */
   /* Actions to execute on various exit codes: */
   struct grecs_list *act_list;
@@ -276,7 +278,11 @@ struct component
   pies_acl_t adm_acl;     /* Administrative ACL (stop, start, etc.) */
 };
 
-#define is_sysvinit(cp) ((cp)->mode >= pies_mark_sysvinit || (cp)->runlevels)
+#define is_sysvinit(cp) \
+	(PIES_SYSVINIT_ENABLED \
+	 && ((cp)->mode >= pies_mark_sysvinit || (cp)->runlevels))
+
+#define SYSVINIT_ACTIVE (PIES_SYSVINIT_ENABLED && init_process)
 
 enum pies_action {
 	ACTION_CONT,
@@ -290,8 +296,6 @@ enum pies_action {
 };
 
 extern char *instance;
-extern char *log_tag;
-extern int log_facility;
 extern unsigned long shutdown_timeout;
 extern struct component default_component;
 extern pies_acl_t pies_acl;
@@ -388,6 +392,8 @@ enum pies_depmap_direction
 
 pies_depmap_t depmap_alloc (size_t count);
 pies_depmap_t depmap_copy (pies_depmap_t dpm);
+size_t depmap_dim (struct pies_depmap *dmap);
+void depmap_free (pies_depmap_t dmap);
 void depmap_set (pies_depmap_t dmap, size_t row, size_t col);
 int depmap_isset (pies_depmap_t dmap, size_t row, size_t col);
 void depmap_clear (pies_depmap_t dmap, size_t row, size_t col);
@@ -399,6 +405,23 @@ size_t depmap_first (pies_depmap_t dmap, enum pies_depmap_direction dir,
 size_t depmap_next (pies_depmap_t dmap, pies_depmap_pos_t pos);
 void depmap_end (pies_depmap_pos_t pos);
 
+
+struct depmap_path_elem
+{
+  int idx;
+  struct depmap_path_elem *next;
+};
+
+struct depmap_path
+{
+  size_t len;
+  struct depmap_path_elem *head, *tail;
+  struct depmap_path *next;
+};
+
+void depmap_path_free (struct depmap_path *path);
+struct depmap_path *depmap_cycle_detect (pies_depmap_t dmap);
+
 int assert_grecs_value_type (grecs_locus_t *locus,
 			     const grecs_value_t *value, int type);
 
@@ -448,11 +471,13 @@ void *register_socket (int fd,
 		       socket_handler_t rd,
 		       socket_handler_t wr,
 		       socket_handler_t ex,
-		       void *data);
+		       void *data,
+		       void (*free_data)(void*));
 void deregister_socket (int fd);
 void update_socket (int fd, int evt, socket_handler_t f);
 
-int register_program_socket (int socktype, int fd, void *data);
+int register_program_socket (int socktype, int fd, void *data,
+			     void (*free_data)(void*));
 int pass_fd (const char *socket, int fd, unsigned time_out);
 int create_socket (struct pies_url *url, int socket_type,
 		   const char *user, mode_t umask);
@@ -497,6 +522,7 @@ void diag_setup (int flags);
 void diagmsg (int logf, int prio, const char *fmt, ...)
   PIES_PRINTFLIKE(3,4);
 
+void vlogmsg (int prio, const char *fmt, va_list ap);
 void logmsg (int prio, const char *fmt, ...) PIES_PRINTFLIKE(2,3);
 void logmsg_printf (int prio, const char *fmt, ...) PIES_PRINTFLIKE(2,3);
 void logmsg_vprintf (int prio, const char *fmt, va_list ap);
@@ -565,6 +591,13 @@ int sysvinit_envlocate (char const *name, char **value);
 int sysvinit_envdelete (char const *name);
 int sysvinit_envupdate (char const *var);
 
+int cb_initdefault (enum grecs_callback_command cmd,
+		    grecs_node_t *node,
+		    void *varptr, void *cb_data);
+int cb_runlevels (enum grecs_callback_command cmd,
+		  grecs_node_t *node,
+		  void *varptr, void *cb_data);
+
 extern char *sysvinit_environ_hint[];
 extern char *init_fifo;
 
