@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2015-2019 Sergey Poznyakoff
+   Copyright (C) 2015-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -744,7 +744,7 @@ shttp_print_error (struct shttp_connection *conn)
 {
   struct json_value *jv;
   
-  if (conn->result && (jv = json_value_lookup (conn->result, "error_message")))
+  if (conn->result && (jv = json_value_lookup (conn->result, "message")))
     {
       if (jv->type == json_string)
 	grecs_error (NULL, 0, "%s", jv->v.s);
@@ -1075,10 +1075,6 @@ print_comp (FILE *fp, struct json_value *v, size_t n)
       else
       	fbuf[fidx++] = '-';
     }
-  else if (strcmp (type, "redirector") == 0)
-    {
-      fbuf[fidx++] = 'R';
-    }
   else if (strcmp (type, "command") == 0)
     {      
       fbuf[fidx++] = 'E';
@@ -1125,12 +1121,6 @@ print_comp (FILE *fp, struct json_value *v, size_t n)
 	    }
 	}
     }
-  else if (strcmp (type, "redirector") == 0)
-    {
-      p = getval (v, "PID", json_number, 0);
-      if (p)
-	fprintf (fp, "%10.0f ", p->v.n);
-    }
   else if (strcmp (type, "command") == 0)
     {
       p = getval (v, "command", json_string, 0);
@@ -1487,7 +1477,7 @@ shttp_print_response_status (struct shttp_connection *conn)
 		    {
 		      if (strcmp (v->v.s, "OK") == 0)
 			fputs (v->v.s, stdout);
-		      else if (json_object_get_type (elt, "error_message",
+		      else if (json_object_get_type (elt, "message",
 						     json_string, &v) == 0)
 			fputs (v->v.s, stdout);
 		      else
@@ -1766,7 +1756,7 @@ telinit_format_environ (struct shttp_connection *conn)
 	  }
 	else if (strcmp (val->v.s, "ER") == 0)
 	  {
-	    if (json_object_get_type (conn->result, "error_message",
+	    if (json_object_get_type (conn->result, "message",
 				      json_string, &val) == 0)
 	      fputs (val->v.s, stderr);
 	    else
@@ -1967,7 +1957,7 @@ conf_reload (struct cmdline_parser_state *state)
   val = json_object_require_type (conn->result, "status", json_string);
   if (strcmp (val->v.s, "ER") == 0)
     {
-      if (json_object_get_type (conn->result, "error_message",
+      if (json_object_get_type (conn->result, "message",
 				json_string, &val) == 0)
 	fputs (val->v.s, stdout);
       else
@@ -2042,7 +2032,7 @@ conf_file_clear (struct cmdline_parser_state *state)
   val = json_object_require_type (conn->result, "status", json_string);
   if (strcmp (val->v.s, "OK"))
     {
-      if (json_object_get_type (conn->result, "error_message",
+      if (json_object_get_type (conn->result, "message",
 				json_string, &val) == 0)
 	fputs (val->v.s, stderr);
       else
@@ -2094,7 +2084,7 @@ conf_file_del (struct cmdline_parser_state *state)
   val = json_object_require_type (conn->result, "status", json_string);
   if (strcmp (val->v.s, "OK") == 0)
     {
-      if (json_object_get_type (conn->result, "error_message",
+      if (json_object_get_type (conn->result, "message",
 				json_string, &val) == 0)
 	{
 	  fputs (val->v.s, stderr);
@@ -2119,7 +2109,7 @@ conf_file_del (struct cmdline_parser_state *state)
     }
   else
     {
-      if (json_object_get_type (conn->result, "error_message",
+      if (json_object_get_type (conn->result, "message",
 				json_string, &val) == 0)
 	fputs (val->v.s, stderr);
       else
