@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2007-2019 Sergey Poznyakoff
+   Copyright (C) 2007-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -94,9 +94,6 @@ prog_tag (struct prog const *prog)
      case TYPE_COMPONENT:
        return prog->v.p.comp->tag;
 
-     case TYPE_REDIRECTOR:
-       return prog->v.r.tag;
-
      case TYPE_COMMAND:
        return prog->v.c.tag;
      }
@@ -147,6 +144,21 @@ unlink_prog (struct prog *pp)
     progtail = pp->prev;
 }
 
+static inline void
+prog_stop_redirectors (struct prog *p)
+{
+  if (p->v.p.redir[RETR_OUT] != -1)
+    {
+      deregister_socket (p->v.p.redir[RETR_OUT]);
+      p->v.p.redir[RETR_OUT] = -1;
+    }
+  if (p->v.p.redir[RETR_ERR] != -1)
+    {
+      deregister_socket (p->v.p.redir[RETR_ERR]);
+      p->v.p.redir[RETR_ERR] = -1;
+    }
+}
+
 void
 destroy_prog (struct prog **pp)
 {
@@ -156,31 +168,14 @@ destroy_prog (struct prog **pp)
   switch (p->type)
     {
     case TYPE_COMPONENT:
+      environ_free (p->v.p.env);
+      if (p->v.p.comp->flags & CF_EXPANDENV)
+	argv_free (p->v.p.argv);
       component_ref_decr (p->v.p.comp);
       if (p->v.p.status == status_listener && p->v.p.socket != -1)
 	deregister_socket (p->v.p.socket);
 	/* FIXME: Remove also all dependent progs (esp. tcpmux) */
-      if (p->v.p.redir[RETR_OUT])
-	p->v.p.redir[RETR_OUT]->v.r.master = NULL;
-      if (p->v.p.redir[RETR_ERR])
-	p->v.p.redir[RETR_ERR]->v.r.master = NULL;
-      break;
-
-    case TYPE_REDIRECTOR:
-      {
-	struct prog *master = p->v.r.master;
-	component_ref_decr (p->v.r.comp);
-	if (master)
-	  {
-	    if (p == master->v.p.redir[0])
-	      master->v.p.redir[0] = NULL;
-	    else if (p == master->v.p.redir[1])
-	      master->v.p.redir[1] = NULL;
-	  }
-	/*	else
-		logmsg (LOG_NOTICE, _("orphan redirector: %s"), p->tag);*/
-	free (p->v.r.tag);
-      }
+      prog_stop_redirectors (p);
       break;
 
     case TYPE_COMMAND:
@@ -191,50 +186,6 @@ destroy_prog (struct prog **pp)
   *pp = NULL;
 }
 
-static char *
-redir_tag (struct prog *master, int type)
-{
-  static char *redirstr[2] = { "stdout", "stderr" };
-  char *str = NULL;
-  size_t len = 0;
-  char const *tag = prog_tag (master);
-  if (type < ARRAY_SIZE(redirstr))
-    grecs_asprintf (&str, &len, "%s/%s", tag, redirstr[type]);
-  else
-    grecs_asprintf (&str, &len, "%s/%d", tag, type);
-  return str;
-}
-
-static struct prog *
-register_redir (int type, struct prog *master)
-{
-  char *tag = redir_tag (master, type);
-  struct prog *pp = grecs_zalloc (sizeof (*pp));
-
-  pp->type = TYPE_REDIRECTOR;
-  pp->v.r.tag = tag;
-  pp->v.r.master = master;
-  pp->v.r.comp = master->v.p.comp;
-  component_ref_incr (pp->v.r.comp);
-  link_prog (pp, NULL);
-  return pp;
-}
-
-void
-update_redir (int type, struct prog *master, pid_t pid)
-{
-  struct prog *pp;
-  if (master->v.p.redir[type])
-    {
-      pp = master->v.p.redir[type];
-      prog_stop (pp, SIGKILL); /* Just in case */
-    }
-
-  pp = register_redir (type, master);
-  master->v.p.redir[type] = pp;
-  pp->pid = pid;
-}
-
 /* Given the component COMP find the element of the program list
    after which to link in the new prog associated with that COMP.
    The progs in the resulting list must be arranged exactly in the
@@ -254,7 +205,7 @@ find_prog_ref (struct component *comp)
     {
       comp = comp->prev;
       if (!comp)
-	return NULL; /* FIXME: Skip redirectors? */
+	return NULL;
     }
 
   if (comp->prog)
@@ -361,12 +312,6 @@ progman_waiting_p (void)
   return 0;
 }
 
-RETSIGTYPE
-redir_exit (int sig)
-{
-  _exit (0);
-}
-
 int
 redirect_to_file (struct prog *master, int stream)
 {
@@ -400,20 +345,113 @@ redirect_to_file (struct prog *master, int stream)
     }
   return fd;
 }
-
-static void
-close_fds (fd_set *fdset)
+
+struct read_buffer
 {
-  int i;
+  struct prog *master;
+  int stream;
+  char text[PIES_LOG_BUF_SIZE];
+  size_t len;
+  int overflow;
+};
 
-  for (i = FD_SETSIZE-1; i >= 0; i--)
+static int
+redirect_read (int fd, void *data)
+{
+  struct read_buffer *rb = data;
+  int n;
+  int prio = rb->master->v.p.comp->redir[rb->stream].v.prio;
+  char const *tag = prog_tag (rb->master);
+  pid_t pid = rb->master->pid;
+  char *p;
+  static char *retr_stream_name[] = {
+    [RETR_OUT] = "stdout",
+    [RETR_ERR] = "stderr"
+  };
+
+  n = read (fd, rb->text + rb->len, sizeof (rb->text) - rb->len);
+  if (n == -1)
     {
-      if (fdset && FD_ISSET (i, fdset))
-	continue;
-      close (i);
+      if (errno != EINTR)
+	{
+	  logmsg (LOG_INFO, _("%s: %s while reading %s"),
+		  tag, strerror (errno), retr_stream_name[rb->stream]);
+	  rb->master->v.p.redir[rb->stream] = -1;
+	  close (fd);
+	  deregister_socket (fd);
+	  return -1;
+	}
     }
+  else if (n == 0)
+    {
+      logmsg (LOG_INFO, _("%s: EOF on %s"),
+	      tag, retr_stream_name[rb->stream]);
+      rb->master->v.p.redir[rb->stream] = -1;
+      close (fd);
+      deregister_socket (fd);
+    }
+  else
+    {
+      rb->len += n;
+
+      while (rb->len)
+	{
+	  p = memchr (rb->text, '\n', rb->len);
+	  if (p)
+	    {
+	      *p++ = 0;
+	      if (rb->overflow)
+		rb->overflow = 0;
+	      else
+		pies_syslog_message (prio, rb->text, tag, pid);
+	      n = rb->len - (p - rb->text);
+	      if (n > 0)
+		memmove (rb->text, p, n);
+	      rb->len = n;
+	    }
+	  else if (rb->len == sizeof (rb->text))
+	    {
+	      rb->text[sizeof (rb->text) - 1] = 0;
+	      pies_syslog_message (prio, rb->text, tag, pid);
+	      rb->len = 0;
+	      rb->overflow = 1;
+	    }
+	  else
+	    break;
+	}
+    }
+  return 0;
 }
 
+static void
+read_buffer_free (void *p)
+{
+  free (p);
+}
+
+int
+redirect_to_syslog (struct prog *master, int stream, int *fd)
+{
+  struct read_buffer *rb;
+  int p[2];
+  
+  if (pipe (p))
+    {
+      logmsg (LOG_CRIT, "pipe: %s", strerror (errno));
+      return -1;
+    }
+
+  rb = grecs_zalloc (sizeof (*rb));
+  rb->master = master;
+  rb->stream = stream;
+  rb->len = 0;
+  rb->overflow = 0;
+  
+  register_socket (p[0], redirect_read, NULL, NULL, rb, read_buffer_free);
+  *fd = p[0];
+  return p[1];
+}
+
 void
 free_redirector (struct redirector *rp)
 {
@@ -422,73 +460,22 @@ free_redirector (struct redirector *rp)
 }
 
 int
-open_redirector (struct prog *master, int stream)
+open_redirector (struct prog *master, int stream, int *fd)
 {
-  int p[2];
-  FILE *fp;
-  char *buf = NULL;
-  size_t size = 0;
-  pid_t pid;
-  int prio;
-  char *tag;
-  fd_set fdset;
-
   switch (master->v.p.comp->redir[stream].type)
     {
     case redir_null:
-      return -1;
-
+      *fd = -1;
+      break;
+      
     case redir_file:
+      *fd = -1;
       return redirect_to_file (master, stream);
 
     case redir_syslog:
-      break;
-    }
-
-  if (pipe (p))
-    {
-      logmsg (LOG_CRIT, "pipe: %s", strerror (errno));
-      return -1;
-    }
-
-  switch (pid = fork ())
-    {
-    case 0:
-      /* Redirector process */
-      tag = redir_tag (master, stream);
-      mf_proctitle_format ("%s redirector", tag);
-      free (tag);
-
-      FD_ZERO (&fdset);
-      FD_SET (p[0], &fdset);
-      close_fds (&fdset);
-
-      diag_setup (0);
-      signal_setup (redir_exit);
-
-      close (p[1]);
-      fp = fdopen (p[0], "r");
-      if (fp == NULL)
-	_exit (1);
-      openlog (prog_tag (master), LOG_PID, master->v.p.comp->facility);
-      prio = master->v.p.comp->redir[stream].v.prio;
-      while (getline (&buf, &size, fp) > 0)
-	syslog (prio, "%s", buf);
-      _exit (0);
-
-    case -1:
-      logmsg (LOG_CRIT,
-		      _("cannot run redirector `%s': fork failed: %s"),
-	      prog_tag (master), strerror (errno));
-      return -1;
-
-    default:
-      debug (1, (_("redirector for %s started, pid=%lu"),
-		 prog_tag (master), (unsigned long) pid));
-      update_redir (stream, master, pid);
-      close (p[0]);
-      return p[1];
+      return redirect_to_syslog (master, stream, fd);
     }
+  return -1;
 }
 
 
@@ -805,6 +792,93 @@ prog_open_socket (struct prog *prog)
   return 0;
 }
 
+static void
+progman_ws_error (const char *fmt, ...)
+{
+  va_list ap;
+
+  va_start (ap, fmt);
+  vlogmsg (LOG_ERR, fmt, ap);
+  va_end (ap);
+}
+
+/* 
+ * Initialization of the environment and argc/argv members of
+ * struct prog.  For regular components this is done in the main
+ * process, so that argv can be reported correctly via the ctl
+ * interface.  For tcpmux components, which are run from the
+ * listener child, this is not so.  That does little harm as
+ * these components are not visible from the ctl interface anyway.
+ */
+static int
+prog_init (struct prog *prog)
+{
+  if ((prog->v.p.env = environ_create (environ)) == NULL)
+    {
+      logmsg (LOG_CRIT, "environ_create: %s", strerror (errno));
+      return -1;
+    }
+  
+  if (prog->v.p.comp->flags & CF_SOCKENV)
+    {
+      size_t i;
+      for (i = 0; sockenv_var[i]; i++)
+	environ_unset (prog->v.p.env, sockenv_var[i], NULL);
+    }
+
+  if (envop_exec (prog->v.p.comp->envop, prog->v.p.env))
+    {
+      logmsg (LOG_CRIT, "environ_exec: %s", strerror (errno));
+      return -1;
+    }
+  
+  if (SYSVINIT_ACTIVE)
+    {
+      size_t i;
+      for (i = 0; sysvinit_environ_hint[i]; i++)
+	{
+	  if (environ_add (prog->v.p.env, sysvinit_environ_hint[i]))
+	    {
+	      logmsg (LOG_CRIT, "environ_add: %s", strerror (errno));
+	      return -1;
+	    }
+	}	      
+    }
+  debug_environ (prog, 4);  
+
+  if (prog->v.p.comp->flags & CF_EXPANDENV)
+    {
+      struct wordsplit ws;
+
+      ws.ws_env = (const char **) environ_ptr (prog->v.p.env);
+      ws.ws_error = progman_ws_error;
+      if (wordsplit (prog->v.p.comp->command, &ws,
+		     WRDSF_QUOTE | WRDSF_SQUEEZE_DELIMS | WRDSF_NOCMD | WRDSF_ENV | WRDSF_ERROR))
+	{
+	  logmsg (LOG_ERR, _("%s: can't split command line: %s"),
+		  prog_tag (prog), wordsplit_strerror (&ws));
+	  return -1;
+	}
+      wordsplit_get_words (&ws, &prog->v.p.argc, &prog->v.p.argv);
+      wordsplit_free (&ws);      
+    }
+  else
+    {
+      prog->v.p.argc = prog->v.p.comp->argc;
+      prog->v.p.argv = prog->v.p.comp->argv;
+    }
+
+  if (debug_level >= 1 && prog->v.p.argv)
+    {
+      int i;
+      logmsg_printf (LOG_DEBUG, "executing");
+      for (i = 0; i < prog->v.p.argc; i++)
+	logmsg_printf (LOG_DEBUG, " %s", quotearg (prog->v.p.argv[i]));
+      logmsg_printf (LOG_DEBUG, "\n");
+    }
+  return 0;
+}
+
 static void
 prog_start_prologue (struct prog *prog)
 {
@@ -816,22 +890,6 @@ prog_start_prologue (struct prog *prog)
 		prog_tag (prog), prog->v.p.comp->dir, strerror (errno));
     }
 
-  prog->v.p.env = environ_create (environ);
-  if (prog->v.p.comp->flags & CF_SOCKENV)
-    {
-      size_t i;
-      for (i = 0; sockenv_var[i]; i++)
-	environ_unset (prog->v.p.env, sockenv_var[i], NULL);
-    }
-  envop_exec (prog->v.p.comp->envop, prog->v.p.env);
-  if (SYSVINIT_ACTIVE)
-    {
-      size_t i;
-      for (i = 0; sysvinit_environ_hint[i]; i++)
-	environ_add (prog->v.p.env, sysvinit_environ_hint[i]);
-    }
-  debug_environ (prog, 4);
-
   pies_priv_setup (&prog->v.p.comp->privs);
   if (prog->v.p.comp->umask)
     umask (prog->v.p.comp->umask);
@@ -839,17 +897,6 @@ prog_start_prologue (struct prog *prog)
   set_limits (prog_tag (prog),
 	      prog->v.p.comp->limits ?
 	      prog->v.p.comp->limits : pies_limits);
-
-  if (debug_level >= 1)
-    {
-      int i;
-      struct component *comp = prog->v.p.comp;
-
-      logmsg_printf (LOG_DEBUG, "executing");
-      for (i = 0; i < comp->argc; i++)
-	logmsg_printf (LOG_DEBUG, " %s", quotearg (comp->argv[i]));
-      logmsg_printf (LOG_DEBUG, "\n");
-    }
 }
 
 static void
@@ -865,14 +912,18 @@ prog_execute (struct prog *prog)
     }
 
   execvp (prog->v.p.comp->program ?
-	  prog->v.p.comp->program : prog->v.p.comp->argv[0],
-	  prog->v.p.comp->argv);
-  openlog (log_tag, LOG_PID, prog->v.p.comp->facility);
+	  prog->v.p.comp->program : prog->v.p.argv[0],
+	  prog->v.p.argv);
+  //FIXME: pies_syslog?  
   syslog (LOG_CRIT, _("cannot start `%s': %s"), prog_tag (prog),
 	  strerror (errno));
   _exit (EX_SOFTWARE);
 }
 
+/*
+ * Run a TCPMUX component.  This function is invoked from a child
+ * process.
+ */
 void
 progman_run_comp (struct component *comp, int fd,
 		  union pies_sockaddr_storage *sa, socklen_t salen)
@@ -882,6 +933,8 @@ progman_run_comp (struct component *comp, int fd,
   prog->v.p.sa_storage = *sa;
   prog->v.p.sa_len = salen;
   prog->v.p.cclass = conn_class_lookup (comp->tag, sa, salen);
+  if (prog_init (prog))
+    _exit (127);
   prog_start_prologue (prog);
   prog_sockenv (prog);
   prog_execute (prog);
@@ -892,7 +945,6 @@ prog_start (struct prog *prog)
 {
   pid_t pid;
   int redir[2];
-  fd_set fdset;
 
   if (prog->pid > 0 || !IS_COMPONENT (prog))
     return;
@@ -972,8 +1024,13 @@ prog_start (struct prog *prog)
       return;
     }
 
-  redir[RETR_OUT] = open_redirector (prog, RETR_OUT);
-  redir[RETR_ERR] = open_redirector (prog, RETR_ERR);
+  if (prog_init (prog))
+    return; //FIXME
+  
+  redir[RETR_OUT] = open_redirector (prog, RETR_OUT,
+				     &prog->v.p.redir[RETR_OUT]);
+  redir[RETR_ERR] = open_redirector (prog, RETR_ERR,
+				     &prog->v.p.redir[RETR_ERR]);
 
   switch (pid = fork ())
     {
@@ -1054,13 +1111,8 @@ prog_start (struct prog *prog)
 	}
 
       /* Close unneeded descripitors */
-      FD_ZERO (&fdset);
-      FD_SET (0, &fdset);
-      FD_SET (1, &fdset);
-      FD_SET (2, &fdset);
-      if (prog->v.p.comp->mode == pies_comp_pass_fd)
-	FD_SET (prog->v.p.socket, &fdset);
-      close_fds (&fdset);
+      pies_close_fds ((prog->v.p.comp->mode == pies_comp_pass_fd
+	               ? prog->v.p.socket : 2) + 1);
 
       prog_execute (prog);
 
@@ -1289,7 +1341,7 @@ prog_create_socket (struct prog *prog, void *data)
 				  comp->privs.user, comp->umask);
 	  if (fd == -1)
 	    destroy_prog (&prog);
-	  else if (register_program_socket (comp->socket_type, fd, prog))
+	  else if (register_program_socket (comp->socket_type, fd, prog, NULL))
 	    {
 	      close (fd);
 	      destroy_prog (&prog);
@@ -1537,15 +1589,6 @@ prog_start_prerequisites (struct prog *prog)
   return ret;
 }
 
-void
-prog_stop_redirectors (struct prog *prog)
-{
-  if (prog->v.p.redir[RETR_OUT])
-    prog_stop (prog->v.p.redir[RETR_OUT], SIGTERM);
-  if (prog->v.p.redir[RETR_ERR])
-    prog_stop (prog->v.p.redir[RETR_ERR], SIGTERM);
-}
-
 void
 prog_stop_dependents (struct prog *prog)
 {
@@ -2004,7 +2047,7 @@ run_command (struct action *act, struct prog *prog, unsigned retcode,
       else
 	setenv ("PIES_STATUS", umaxtostr (STATUS_CODE (retcode), buf), 1);
 
-      close_fds (NULL);
+      pies_close_fds (3);
 
       argv[0] = "/bin/sh";
       argv[1] = "-c";
@@ -2182,17 +2225,6 @@ progman_cleanup (int expect_term)
 
 	  break;
 
-	case TYPE_REDIRECTOR:
-	  /* It was a redirector of an already finished process. */
-	  print_status (prog_tag (prog), pid, status,
-			expect_term ||
-			(prog->v.r.master &&
-			 prog->v.r.master->v.p.status == status_stopping));
-	  debug (1, (_("removing redirector %s, pid=%lu"),
-		     prog_tag (prog), (unsigned long)pid));
-	  destroy_prog (&prog);
-	  break;
-
 	case TYPE_COMMAND:
 	  print_status (prog_tag (prog), pid, status, expect_term);
 	  destroy_prog (&prog);
@@ -2279,7 +2311,7 @@ prog_activate_listener (struct prog *prog)
 			      comp->privs.user, comp->umask);
       if (fd == -1)
 	return -1;
-      else if (register_program_socket (comp->socket_type, fd, prog))
+      else if (register_program_socket (comp->socket_type, fd, prog, NULL))
 	{
 	  close (fd);
 	  return -1;
