@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2013-2019 Sergey Poznyakoff
+   Copyright (C) 2013-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -625,7 +625,7 @@ create_fifo (void)
               strerror (errno));
       return;
     }
-  register_socket (fd, sysvinit_fifo_handler, NULL, NULL, NULL);
+  register_socket (fd, sysvinit_fifo_handler, NULL, NULL, NULL, NULL);
 }
 
 static char *try_console[] = { NULL, "/dev/console", "/dev/tty0" };
@@ -1167,7 +1167,7 @@ inittab_parse (const char *file)
   fp = fopen (file, "r");
   if (fp)
     {
-      while (getline (&ctx.buf, &ctx.size, fp) >= 0)
+      while (grecs_getline (&ctx.buf, &ctx.size, fp) >= 0)
 	{
 	  enum inittab_status st = inittab_parse_line (&ctx);
 	  if (st == inittab_err)
